
/* 
 * Recursive Binary Inertial Partition
 *
 * Purpose
 *   Divides the object perpendicular to the long axis of the object
 *   as determined by the relative magnitude of the moment of inertia
 *   along each principal axis.
 *
 * Input
 *   ndata is the number of elements (nodes)
 *   x[] is the list of X coordinates
 *   y[] is the list of Y coordinates
 *   z[] is the list of Z coordinates
 *   level is the level of recursion, i.e. the domain will be
 *	divided into (2 ^ level) subdomains
 *
 * Output
 *   no output is returned
 *
 * Note
 *   coordinates x[], y[], and z[] are redistributed based on
 *	the result of binary inertial partitioning
 */

#include <stdio.h>
#include <math.h>
#include "parti.h"
#include "bisec.h"

#define  MAXPART	512	/* maximum number of partitions */
#define	 NDIM		4	/* 3-dimenstional */
#define	 NTENSOR	10	/* size of each inertial tensor
				   (3x3 symmetric matrix) */
#define	 XX		0	/* moment of inertial about the x axis */
#define	 XY		1	/* xy product of inertial */
#define	 XZ		2	/* zx product of inertial */
#define	 YY		3	/* moment of inertial about the y axis */
#define	 YZ		4	/* yz product of inertial */
#define	 ZZ		5	/* moment of inertial about the z axis */

extern	double *doublelist;	/* array of double numbers */

void
RecInerBisec (level, remaplevel, myindex, ndata, ndim, data)
int    level;			/* level of binary disection */
int    remaplevel;		/* remap for every remaplevel */
int    *myindex;		/* current indexes local data */
int    *ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
  int    curndata;		/* cursnt size */
  int    totalpart;		/* total number of partitions */
  int    npart;			/* current number of partitions */
  int    parlevel;		/* level of each partition */
  int    parproc;		/* # of processors of each partition */
  int    mask;			/* mask */
  int    ptr[MAXPART+1];	/* pointer of begin of partition */
  int    gsize[MAXPART];	/* size of each partition */
  int    work[MAXPART];		/* working array */
  int    *proc;			/* processor assignments */
  int    *index;		/* index of arrays */
  int    *offset;		/* offset within partitions */
  int    i, j, k, l, m, n, no;
  double C[MAXPART][NDIM];	/* array of centers of masses */
  double V[MAXPART][NDIM];	/* array of eigenvectors */
  double M[MAXPART][NTENSOR];	/* array of inertial tensors */
  double dwork[MAXPART][NTENSOR];/* double working array */
  double *dot;			/* array of inner product */
  Val    *midlist;		/* array of medians */
  Val    *temp;
  SCHED  *sched;		/* schedule for remapping */
  TTABLE *table;		/* translation table */
  TTABLE *init_ttable_with_proc();
  int	 comparedouble();

  /* check the remapping level */
  if (remaplevel <= 0) {
    printf ("Remapping level must be greater than 0.\n");
    exit(1);
  }

  /* initialize parameters */
  curndata = *ndata;
  totalpart = (1 << remaplevel);

  /* allocate memory for necessary arrays*/
  index   = (int *) malloc ((curndata+PARTI_numnodes) * sizeof(int));
  proc    = (int *) malloc ((curndata+PARTI_numnodes) * sizeof(int));
  dot     = (double *) malloc ((curndata+PARTI_numnodes) * sizeof(double));
  offset  = (int *) malloc ((PARTI_numnodes) * sizeof(int));
  midlist = (Val *) malloc ((totalpart * PARTI_numnodes) * sizeof(Val));
  temp    = (Val *) malloc ((totalpart * PARTI_numnodes) * sizeof(Val));
  doublelist  = dot;

  /* put the coordinates into data array */
  for (i = 0; i <= curndata; i++) {
    index[i] = i;
    proc[i] = 0;
  }

  /* sort the arrays of each partition and find the median */
  npart = 1;				/* number of partitions */
  for (i=0, ptr[0]=0, parlevel=level; i < level; i++) {
    /* remap all data if level is equal to remaplevel */
    if ((i / remaplevel) && ((i % remaplevel) == FALSE)) {
      /* calculate the number of processors of each partition */
      parlevel -= remaplevel;		/* level for global func */
      parproc = 1 << parlevel;		/* # of processors */
      mask = parproc - 1;		/* mask for proc */

      /* set an offset array for distribution */
      for (j = 0; j < npart; j++)
        offset[j] = PARTI_mynode & mask;

      /* set up the proc array */
      for (j = 0; j < curndata; j++) {
    	offset[proc[j]] &= mask;
    	proc[j] = (proc[j] << parlevel) + offset[proc[j]]++;
      }

      /* output the mapping array
      for (j = 0; j < PARTI_numnodes; j++) {
        if (PARTI_mynode == j) {
    	  pfd = fopen ("map", "a+");
    	  for (k = 0; k < curndata; k++)
    	    fprintf(pfd, "%d\n", proc[k]);
    	  fclose (pfd);
        }
        gsync();
      }	*/

      /* build translation table */
      table = init_ttable_with_proc(1, proc, curndata);

      /* call remap function to build a schedule */
      remap (table, myindex, &sched, myindex, &curndata);

      /* move the x, y, and z arrays */
      for (j = 0; j < ndim; j++)
        PARTI_gather (sched, data[j], data[j], sizeof(double));

      /* free the schedule */
      free_sched (sched);
      free_table (table);

      /* reset the proc array */
      for (j = 0; j <= curndata; j++) {
    	proc[j] = PARTI_mynode >> parlevel;
    	index[j] = j;
      }

      /* number of partition per subcube */
      npart = 1;
    }

    /* find the global sizes of partitions */
    ptr[npart] = curndata;    		/* number of nodes */
    for (j = 0; j < npart; j++)
      gsize[j] = ptr[j+1] - ptr[j];
    addivec (parlevel, gsize, npart, work);

    /* find the center of mass of each partition */
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
        C[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (m = 0; m < ndim; m++)
          C[j][m] += data[m][n];
      }
    }
    adddvec (parlevel, C, npart*NDIM, dwork);
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
    	C[j][k] /= gsize[j];

#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Center of mass: %f %f %f\n", C[0][0], C[0][1], C[0][2]);
#endif

    /* calculate the moment of inertial tensor */
    for (j = 0; j < npart; j++)
      for (k = 0; k < NTENSOR; k++)
    	M[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (no = 0, l = 0; l < ndim; l++)
	  for (m = l; m < ndim; no++, m++)
            M[j][no] += (data[l][n] - C[j][l]) * (data[m][n] - C[j][m]);
      }
    }
    adddvec (parlevel, M, npart*NTENSOR, dwork);
#ifdef	DEBUG
    if (PARTI_mynode == 0) {
      printf ("Moment of Inertial Tensor:\n");
      printf ("\t%f\t%f\t%f\n", M[0][XX], M[0][XY], M[0][XZ]);
      printf ("\t%f\t%f\t%f\n", M[0][XY], M[0][YY], M[0][YZ]);
      printf ("\t%f\t%f\t%f\n", M[0][XZ], M[0][YZ], M[0][ZZ]);
    }
#endif

    /* find the eigenvalues of eigenvectors */
    for (j = 0; j < npart; j++)
      find_eigenvector (ndim, &M[j][0], &V[j][0]);

#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Eigenvector: %f %f %f\n", V[0][0], V[0][1], V[0][2]);
#endif

    /* projection (inner product) of each element to eigenvector */
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
        n = index[k];
	for (dot[n] = 0, m = 0; m < ndim; m++)
          dot[n] += V[j][m] * (data[m][n] - C[j][m]);
      }
    }

    /* sort the dot product array of each partition */
    for (j = 0; j < npart; j++)
      qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);

    /* apply the method median of medians to find 
       the median of the distributed array */
    find_median (parlevel, npart, ptr, dot, index, proc, midlist, temp);

    /* number of partitions */
    npart <<= 1;
  }

  /* build translation table */
  table = init_ttable_with_proc(1, proc, curndata);

  /* call remap function to build a schedule */
  remap (table, myindex, &sched, myindex, ndata);

  /* move the x, y, and z arrays */
  for (j = 0; j < ndim; j++)
    PARTI_gather (sched, data[j], data[j], sizeof(double));

  /* free the schedule */
  free_sched (sched);
  free_table (table);

  /* release memory */
  free (index);
  free (proc);
  free (ptr);
  free (dot);
  free (offset);
  free (temp);
  free (midlist);
}

void
WeighRecInerBisec (level, remaplevel, myindex, load, ndata, ndim, data)
int    level;			/* level of binary disection */
int    remaplevel;		/* remap for every remaplevel */
int    *myindex;		/* current indexes local data */
int    *load;			/* load of each element */
int    *ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
  int    curndata;		/* cursnt size */
  int    totalpart;		/* total number of partitions */
  int    npart;			/* current number of partitions */
  int    parlevel;		/* level of each partition */
  int    parproc;		/* # of processors of each partition */
  int    mask;			/* mask */
  int    ptr[MAXPART+1];	/* pointer of begin of partition */
  int    gsize[MAXPART];	/* size of each partition */
  int    work[MAXPART];		/* working array */
  int    *proc;			/* processor assignments */
  int    *index;		/* index of arrays */
  int    *offset;		/* offset within partitions */
  int    *weight;		/* weight array */
  int    i, j, k, l, m, n, no;
  double C[MAXPART][NDIM];	/* array of centers of masses */
  double V[MAXPART][NDIM];	/* array of eigenvectors */
  double M[MAXPART][NTENSOR];	/* array of inertial tensors */
  double dwork[MAXPART][NTENSOR];/* double working array */
  double *dot;			/* array of inner product */
  Val    *midlist;		/* array of medians */
  Val    *temp;
  SCHED  *sched;		/* schedule for remapping */
  TTABLE *table;		/* translation table */
  TTABLE *init_ttable_with_proc();
  int	 comparedouble();

  /* check the remapping level */
  if (remaplevel <= 0) {
    printf ("Remapping level must be greater than 0.\n");
    exit(1);
  }

  /* initialize parameters */
  curndata = *ndata;
  totalpart = (1 << remaplevel);

  /* allocate memory for necessary arrays*/
  index   = (int *) malloc ((curndata<<1) * sizeof(int));
  proc    = (int *) malloc ((curndata<<1) * sizeof(int));
  dot     = (double *) malloc ((curndata<<1) * sizeof(double));
  weight  = (int *) malloc ((curndata<<1) * sizeof(int));
  offset  = (int *) malloc (PARTI_numnodes * sizeof(int));
  midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
  temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
  doublelist  = dot;

  /* put the coordinates into data array */
  for (i = 0; i <= curndata; i++) {
    index[i] = i;
    proc[i] = 0;
  }

  /* sort the arrays of each partition and find the median */
  npart = 1;				/* number of partitions */
  for (i=0, ptr[0]=0, parlevel=level; i < level; i++) {
    /* remap all data if level is equal to remaplevel */
    if ((i / remaplevel) && ((i % remaplevel) == FALSE)) {
      /* calculate the number of processors of each partition */
      parlevel -= remaplevel;		/* level for global func */
      parproc = 1 << parlevel;		/* # of processors */
      mask = parproc - 1;		/* mask for proc */

      /* set an offset array for distribution */
      for (j = 0; j < npart; j++)
        offset[j] = PARTI_mynode & mask;

      /* set up the proc array */
      for (j = 0; j < curndata; j++) {
    	offset[proc[j]] &= mask;
    	proc[j] = (proc[j] << parlevel) + offset[proc[j]]++;
      }

      /* output the mapping array
      for (j = 0; j < PARTI_numnodes; j++) {
        if (PARTI_mynode == j) {
    	  pfd = fopen ("map", "a+");
    	  for (k = 0; k < curndata; k++)
    	    fprintf(pfd, "%d\n", proc[k]);
    	  fclose (pfd);
        }
        gsync();
      }	*/

      /* build translation table */
      table = init_ttable_with_proc(1, proc, curndata);

      /* call remap function to build a schedule */
      remap (table, myindex, &sched, myindex, &curndata);

      /* move the x, y, and z arrays */
      for (j = 0; j < ndim; j++)
        PARTI_gather (sched, data[j], data[j], sizeof(double));
      PARTI_gather (sched, load, load, sizeof(int));

      /* free the schedule */
      free_sched (sched);
      free_table (table);

      /* reset the proc array */
      for (j = 0; j <= curndata; j++) {
    	proc[j] = PARTI_mynode >> parlevel;
    	index[j] = j;
      }

      /* number of partition per subcube */
      npart = 1;
    }

    /* find the mass of partitions */
    ptr[npart] = curndata;    		/* number of nodes */
    for (j = 1, weight[0] = 0; j <= curndata; j++)
      weight[j] = weight[j-1] + load[index[j-1]];
    for (j = 0; j < npart; j++)
      gsize[j] = weight[ptr[j+1]] - weight[ptr[j]];
    addivec (parlevel, gsize, npart, work);

    /* find the center of mass of each partition */
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
        C[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (m = 0; m < ndim; m++)
          C[j][m] += load[n] * data[m][n];
      }
    }
    adddvec (parlevel, C, npart*NDIM, dwork);
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
    	C[j][k] /= gsize[j];

    /* calculate the moment of inertial tensor */
    for (j = 0; j < npart; j++)
      for (k = 0; k < NTENSOR; k++)
    	M[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (no = 0, l = 0; l < ndim; l++)
	  for (m = l; m < ndim; no++, m++)
            M[j][no] += load[n] * (data[l][n]-C[j][l]) * (data[m][n]-C[j][m]);
      }
    }
    adddvec (parlevel, M, npart*NTENSOR, dwork);

    /* find the eigenvalues of eigenvectors */
    for (j = 0; j < npart; j++)
      find_eigenvector (ndim, &M[j][0], &V[j][0]);

    /* projection (inner product) of each element to eigenvector */
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
        n = index[k];
	for (dot[n] = 0, m = 0; m < ndim; m++)
          dot[n] += V[j][m] * (data[m][n] - C[j][m]);
      }
    }

    /* sort the dot product array of each partition */
    for (j = 0; j < npart; j++)
      qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);

    /* calculate the weight */
    for (j = 1, weight[0] = 0; j <= curndata; j++)
      weight[j] = weight[j-1] + load[index[j-1]];

    /* apply the method median of medians to find 
       the median of the distributed array */
    find_median_weight (parlevel, npart, ptr, dot, weight, index, proc, midlist, temp);

    /* number of partitions */
    npart <<= 1;
  }

  /* build translation table */
  table = init_ttable_with_proc(1, proc, curndata);

  /* call remap function to build a schedule */
  remap (table, myindex, &sched, myindex, ndata);

  /* move the x, y, and z arrays */
  for (j = 0; j < ndim; j++)
    PARTI_gather (sched, data[j], data[j], sizeof(double));
  PARTI_gather (sched, load, load, sizeof(int));

  /* restore the load array
  for (i = 1; i < *ndata; i++)
    load[i] += load[i-1]; */

  /* free the schedule */
  free_sched (sched);
  free_table (table);

  /* release memory */
  free (index);
  free (proc);
  free (ptr);
  free (dot);
  free (weight);
  free (offset);
  free (temp);
  free (midlist);
}
